@@ -1,5 +1,5 @@
 /*  GNU ddrescue - Data recovery tool
-    Copyright (C) 2004, 2005 Antonio Diaz Diaz.
+    Copyright (C) 2004, 2005, 2006 Antonio Diaz Diaz.
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -26,6 +26,7 @@
 #include <cstdlib>
 #include <vector>
 #include <unistd.h>
+
 #include "ddrescue.h"
 
 
@@ -82,8 +83,8 @@ void compact_sblock_vector( std::vector< Sblock > & sblock_vector ) throw()
   }
 
 
-void complete_sblock_vector( std::vector< Sblock > & sblock_vector,
-                             const long long isize ) throw()
+void extend_sblock_vector( std::vector< Sblock > & sblock_vector,
+                           const long long isize ) throw()
   {
   if( sblock_vector.size() == 0 )
     {
@@ -164,12 +165,13 @@ void split_domain_border_sblocks( std::vector< Sblock > & sblock_vector,
 void logfile_error( const char * filename, const int linenum ) throw()
   {
   char buf[80];
-  std::snprintf( buf, sizeof( buf ), "error in logfile %s, line %d\n",
+  snprintf( buf, sizeof( buf ), "error in logfile %s, line %d\n",
                  filename, linenum );
   show_error( buf );
   }
 
 
+// Writes periodically the logfile to disc.
 // Returns false only if update is attempted and fails.
 //
 bool update_logfile( const std::vector< Sblock > & sblock_vector,
@@ -179,16 +181,17 @@ bool update_logfile( const std::vector< Sblock > & sblock_vector,
   static time_t t1 = std::time( 0 );
 
   if( !filename ) return true;
+  const int interval = 30 + std::min( 270, (int)sblock_vector.size() / 40 );
   const time_t t2 = std::time( 0 );
-  if( force || t2 - t1 >= 30 ) t1 = t2;		// update every 30 seconds
-  else return true;
+  if( !force && t2 - t1 < interval ) return true;
+  t1 = t2;
   fsync( odes );
 
   FILE *f = std::fopen( filename, "w" );
   if( !f )
     {
     char buf[80];
-    std::snprintf( buf, sizeof( buf ), "error opening logfile %s for writing", filename );
+    snprintf( buf, sizeof( buf ), "error opening logfile %s for writing", filename );
     show_error( buf, errno );
     return false;
     }
@@ -206,7 +209,7 @@ bool update_logfile( const std::vector< Sblock > & sblock_vector,
   if( std::fclose( f ) )
     {
     char buf[80];
-    std::snprintf( buf, sizeof( buf ), "error writing logfile %s", filename );
+    snprintf( buf, sizeof( buf ), "error writing logfile %s", filename );
     show_error( buf, errno );
     return false;
     }
@@ -227,9 +230,9 @@ void Logbook::set_rescue_domain( const long long ipos, const long long opos,
     if( _domain.pos() >= isize )
       {
       char buf[80];
-      std::snprintf( buf, sizeof( buf ), "can't start reading at pos %lld",
+      snprintf( buf, sizeof( buf ), "can't start reading at pos %lld",
                      _domain.pos() ); show_error( buf );
-      std::snprintf( buf, sizeof( buf ), "input file is only %lld bytes long",
+      snprintf( buf, sizeof( buf ), "input file is only %lld bytes long",
                      isize ); show_error( buf );
       std::exit(1);
       }
@@ -386,7 +389,7 @@ int Logbook::copy_errors() throw()
     const int msglen = std::strlen( msgbuf );
     for( int retry = 1; _max_retries < 0 || retry <= _max_retries; ++retry )
       {
-      std::snprintf( msgbuf + msglen, sizeof( msgbuf ) - msglen, "%d", retry );
+      snprintf( msgbuf + msglen, sizeof( msgbuf ) - msglen, "%d", retry );
       unsigned int index = 0;
       bool first_post = true, bad_block_found = false;
       split_domain_border_sblocks( sblock_vector, _domain );
@@ -425,8 +428,11 @@ int Logbook::copy_errors() throw()
             if( index > 0 && sblock_vector[index-1].join( result[0] ) )
               result.erase( result.begin() );
             if( result.size() )
-              { sblock_vector.insert( sblock_vector.begin() + index, result.begin(), result.end() );
-              index += result.size(); }
+              {
+              sblock_vector.insert( sblock_vector.begin() + index,
+                                    result.begin(), result.end() );
+              index += result.size();
+              }
             }
           if( retval ) return retval;
           if( !update_logfile( sblock_vector, filename, _odes ) ) return 1;
@@ -442,8 +448,8 @@ int Logbook::copy_errors() throw()
 Logbook::Logbook( const long long ipos, const long long opos,
                   const long long max_size, const long long isize,
                   const char * name, const int cluster, const int hardbs,
-                  const int max_errors, const int max_retries,
-                  const int verbosity, const bool nosplit ) throw()
+                  const int max_errors, const int max_retries, const int verbosity,
+                  const bool complete_only, const bool nosplit ) throw()
   : filename( name ), _hardbs( hardbs ), _softbs( cluster * hardbs ),
     _max_errors( max_errors ), _max_retries( max_retries ),
     _verbosity( verbosity ), _nosplit( nosplit )
@@ -460,7 +466,19 @@ Logbook::Logbook( const long long ipos, const long long opos,
 
   set_rescue_domain( ipos, opos, max_size, isize );
   if( filename ) read_logfile();
-  complete_sblock_vector( sblock_vector, isize );
+  if( !complete_only ) extend_sblock_vector( sblock_vector, isize );
+  else		// limit _domain to blocks of finite size read from logfile
+    {
+    if( sblock_vector.size() && sblock_vector.back().size() < 0 )
+      sblock_vector.pop_back();
+    if( sblock_vector.size() )
+      {
+      const Block b( sblock_vector.front().pos(),
+                     sblock_vector.back().end() - sblock_vector.front().pos() );
+      _domain = b.overlap( _domain );
+      }
+    else _domain.size( 0 );
+    }
   compact_sblock_vector( sblock_vector );
   }
 
@@ -479,8 +497,8 @@ int Logbook::do_rescue( const int ides, const int odes ) throw()
 
   for( unsigned i = 0; i < sblock_vector.size(); ++i )
     {
-    Sblock & sb = sblock_vector[i];
-    Block b = sb.overlap( _domain );
+    const Sblock & sb = sblock_vector[i];
+    const Block b = sb.overlap( _domain );
     if( b.size() == 0 ) { if( sb < _domain ) continue; else break; }
     switch( sb.status() )
       {
@@ -498,9 +516,8 @@ int Logbook::do_rescue( const int ides, const int odes ) throw()
     if( filename )
       {
       std::printf( "Initial status (read from logfile)\n" );
-      std::printf( "rescued: %10sB,", format_num( recsize ) );
-      std::printf( "  errsize:%9sB,", format_num( errsize, 99999 ) );
-      std::printf( "  errors: %7u\n", errors );
+      std::printf( "rescued: %10sB,  errsize:%9sB,  errors: %7u\n",
+                   format_num( recsize ), format_num( errsize, 99999 ), errors );
       std::printf( "Current status\n" );
       }
     }
